/*
Nathan Monroe
5/10/2013
6.UAP Final Project
MIDI Guitar
main.c
*/
#include <device.h>
#include "Key.h"

#define MAX_LCD_COL             16
#define ON (0x01u) //MIDI on and off
#define OFF (0x00u)

/* Identity Reply message */
const uint8 CYCODE MIDI_IDENTITY_REPLY[] = {
    0xF0u,      /* SysEx */
    0x7Eu,      /* Non-Realtime */
    0x7Fu,      /* ID of target device (7F - "All Call") */
    0x06u,      /* Sub-ID#1 - General Information */
    0x02u,      /* Sub-ID#2 - Identity Reply */
    0x7Du,      /* Manufacturer's ID: 7D - Educational Use */
    0xB4u, 0x04u,               /* Family code */
    0x32u, 0xD2u,               /* Model number */
    0x01u, 0x00u, 0x00u, 0x00u, /* Version number */
    /*0xF7         End of SysEx automatically appended */
};

/* Need for Identity Reply message */
extern volatile uint8 USB_MIDI1_InqFlags;
extern volatile uint8 USB_MIDI2_InqFlags;

volatile uint8 usbActivityCounter = 0u;
uint8 midiMsg[4]; //Midi message to be sent to PC
int16 string_absmax[6]; // absolute maximum value for strings. String [0]-[5] maps to string 1-6.
uint16 string_absmin[6] = {1000u, 1000u, 1000u, 1000u, 1000u, 1000u}; //Absolute minimum value for strings. Used to get rid of steady state error from ADC.

uint16 string_samples[6][100]; //current string value stores previous 100 samples (~1.5 periods at 5.8KHz sampling, 82.4Hz minimum freq.
uint16 this_sample; //Current Sample from ADC
uint8 string_tocheck; //Next string to be read from ADC
uint8 sample_number = 0u; //Number of sample to be read into string_samples
uint16 curr_max; //current maximum value of string amplitude
uint8 string_active[6]; //tells if the current string is playing
uint8 fret_active[6]; //fret number of active string
uint8 thismidinote; //MIDI note to be sent out
uint8 checked_strings[6] = {0, 0, 0, 0, 0, 0}; //strings that are being checked for play activity
uint8 t = 0; //for iterating
uint8 num_high_strings = 0; //number of strings with a high voltage
uint8 returned_string;
uint8 checked_string;
uint8 display_note; //note to display on LCD
uint16 off_thres; //Threshold for MIDI off events
uint8 delay[6]; //used for delaying analog measurements
uint8 midi_off_note; //used to prevent same-string collisions
extern uint8 const CYCODE LCD_customFonts[];

uint8 Check_Fret() { //Returns Fret number of what's being played. Higher frets have priority, corresponding to guitar physical design.
if (Fret_12_Read()) return 12;
else {
    if (Fret_11_Read()) return 11;
    else {
        if (Fret_10_Read()) return 10;
        else {
            if (Fret_9_Read()) return 9;
            else {
                if (Fret_8_Read()) return 8;
                else {
                    if (Fret_7_Read()) return 7;
                    else {
                        if (Fret_6_Read()) return 6;
                        else {
                            if (Fret_5_Read()) return 5;
                            else {
                                if (Fret_4_Read()) return 4;
                                else {
                                    if (Fret_3_Read()) return 3;
                                    else {
                                        if (Fret_2_Read()) return 2;
                                        else {
                                            if (Fret_1_Read()) return 1;
                                            else return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}

}


void adc_change_string(uint8 string) { //Strings index 0 to 5. 
    ADC_DelSig_StopConvert(); //Don't want garbage coming out from switching during a conversion.
    AMux_1_FastSelect(string);
    ADC_DelSig_StartConvert();
}



uint16 curr_stringmax(uint8 string){ //Returns current maximum value of a string over the last 1.5 periods.
    curr_max = 0;
    int x;
    for (x=0; x<100; x++){
        if (string_samples[string][x] > curr_max) curr_max = string_samples[string][x];
    }
    return abs(curr_max) - string_absmin[string]; //Normalize for zero-mean

}
void read_next_string() {
    if(ADC_DelSig_IsEndConversion(ADC_DelSig_RETURN_STATUS)) //Check if ADC is done
    {   
        ADC_DelSig_StopConvert();
        this_sample = ADC_DelSig_GetResult16();
        if (!calib_Read()) { //If the calibration button is being asserted
            string_absmin[string_tocheck] = ((curr_stringmax(string_tocheck)*3)>>2)+ string_absmin[string_tocheck]; //Reset thresholds
                                                                                //absmin is set to 3/4 measured value to prevent underflow
            string_absmax[0] = 10u; //These values will be increased upon picking each string
            string_absmax[1] = 10u;
            string_absmax[2] = 10u;
            string_absmax[3] = 10u;
            string_absmax[4] = 10u;
            string_absmax[5] = 10u;
        }
        else{
            string_samples[string_tocheck][sample_number] = this_sample; //put sample into array
            if ((this_sample < string_absmin[string_tocheck])) string_absmin[string_tocheck] = this_sample  - 10u; //If it overflowed negatively, increase the string minimum so it doesn't overflow. This can really mess up note detection.
            //if ((string_active[string_tocheck] == 0x0) & ((this_sample - string_absmin[string_tocheck]) > 15u)) string_absmin[string_tocheck] = this_sample  - 10u;

            if (string_samples[string_tocheck][sample_number] - string_absmin[string_tocheck] > string_absmax[string_tocheck]) string_absmax[string_tocheck] = string_samples[string_tocheck][sample_number] - string_absmin[string_tocheck]; //If a new absolute maximum sample has been found, update the maximum array accordingly.       
        }
        
        string_tocheck = string_tocheck + 1; //Go to the next string
        if (string_tocheck > 5u) { //Mod 5
            string_tocheck = 0u;
            sample_number = sample_number + 1; //Go to the next sample
            if (sample_number > 99u) sample_number = 0; //Mod 100
        }
        adc_change_string(string_tocheck);
    }
}
uint8 Check_String() { //Returns string number of what's being played. returns 1-6 corresponding to standard string number convention.
    checked_strings[0] = 0u; //Reset values
    checked_strings[1] = 0u;
    checked_strings[2] = 0u;
    checked_strings[3] = 0u;
    checked_strings[4] = 0u;
    checked_strings[5] = 0u;
    num_high_strings = 0;
    returned_string = 0;
    if (String_1_Read()) checked_strings[0] = 0x1;
    if (String_2_Read()) checked_strings[1] = 0x1;
    if (String_3_Read()) checked_strings[2] = 0x1;
    if (String_4_Read()) checked_strings[3] = 0x1;
    if (String_5_Read()) checked_strings[4] = 0x1;
    if (String_6_Read()) checked_strings[5] = 0x1;
   
    for (t=0; t<6; t++) { //iterate over strings that are high
        if (checked_strings[t] == 0x1) num_high_strings = num_high_strings + 1;
    }
    if (num_high_strings == 0u) returned_string = 0; //no strings scenario
    else if (num_high_strings == 1u) { //1 string scenario
        for (t=0; t<6; t++) {
            if (checked_strings[t] == 0x1) returned_string = t+1;
        }
    }
    else { //multiple strings scenario, collision detection and decision making
        CyDelay(5); //5ms ~half period
        read_next_string(); //Sample all 6 strings once
        CyDelayUs(200);
        read_next_string(); //2
        CyDelayUs(200);
        read_next_string(); //3
        CyDelayUs(200);
        read_next_string(); //4
        CyDelayUs(200);
        read_next_string(); //5
        CyDelayUs(200);
        read_next_string(); //6
        CyDelayUs(200);
        if (String_6_Read()){
            returned_string = 6;
        }
        
        //Return the string with highest normalized reading
        //Divide absmx over curr_max to save having to deal with floating point calculations
        if (String_5_Read() & ((string_absmax[4]/curr_stringmax(4)) < (string_absmax[returned_string-1]/curr_stringmax(returned_string-1)))) returned_string = 5;
        if (String_4_Read() & ((string_absmax[3]/curr_stringmax(3)) < (string_absmax[returned_string-1]/curr_stringmax(returned_string-1)))) returned_string = 4;
        if (String_3_Read() & ((string_absmax[2]/curr_stringmax(2)) < (string_absmax[returned_string-1]/curr_stringmax(returned_string-1)))) returned_string = 3;
        if (String_2_Read() & ((string_absmax[1]/curr_stringmax(1)) < (string_absmax[returned_string-1]/curr_stringmax(returned_string-1)))) returned_string = 2;
        if (String_1_Read() & ((string_absmax[0]/curr_stringmax(0)) < (string_absmax[returned_string-1]/curr_stringmax(returned_string-1)))) returned_string = 1;
    
    }
    return returned_string;
    
}

void sendMIDI(uint8 note, uint8 on) {
    if(USB_GetConfiguration() != 0u)    // Service USB MIDI when device configured
    {
        midiMsg[0] = USB_MIDI_NOTE_ON;
        if (on == ON) {
            midiMsg[2] = 100u; //MidiMsg[2] is note velocity. 
        }
        else {
            midiMsg[2] = 0u; //Velocity of 0 is equivalent to note turning off.
        }
        midiMsg[1] = note; //midiMsg[1] is the code of the note that's being changed.
        USB_PutUsbMidiIn(3u, midiMsg, USB_MIDI_CABLE_00); //Send the message to the MIDI output buffer
        USB_MIDI_IN_Service(); //Service the buffer.       
    }
}

void main()
{
    int y; //for iteration
    uint8 prevbuffer2;
    uint8 buffer[3]; //I2C Buffer
                     //buffer[0] is string, 
                     //buffer[1] is fret, 
                     //buffer[2] is string being picked
    buffer[0] = 0; //Initialize buffer to 0
    buffer[1] = 0;
    buffer[2] = 0;
    //buffer[3] = 0;

    LCD_Start(); //Initialize LCD
    
    AMux_1_Init(); //Initialize Mux
    AMux_1_FastSelect(0);  
    
    PGA_Init(); //Initialize Gain amplifier
    PGA_Start();
    
    EZI2C_1_Start(); //Initialize I2C Slave
    EZI2C_1_SetBuffer1(3,1,(void *)(buffer));
    
    ADC_DelSig_Start(); //Initialize ADC and start first conversion
    ADC_DelSig_StartConvert();

    CYGlobalIntEnable;  //Enable Interrupts
    USB_Start(0u, USB_DWR_VDDD_OPERATION);  //Start MIDI USB Slave

       
    while(1)
    {
//////////////////////I2C Handling
         read_next_string();
         checked_string = Check_String(); //Do this just once per cycle to improve performance during multi-string scenarios
         buffer[0] = checked_string; //Put String and Fret information into I2C buffer
         buffer[1] = Check_Fret();
         if (checked_string != 0u) buffer[2] = 0x1; //Put note on information into I2C buffer
         else {
            buffer[2] = 0x0;
         }
//////////////////////MIDI Note off Events
        for (y=1; y<7; y++){ //Go through all the strings
            off_thres = 0; //Dynamically update threshold for 'off' event based on fret status
            if (string_active[y-1]==0x1) {
                if (fret_active > 0) off_thres = (string_absmax[y-1] >> 1);
                else off_thres = (string_absmax[y-1] >> 2); //Fretted notes vibrate with a lower amplitude
                delay[y-1] = delay[y-1] - 1; 
                if (delay[y-1] == 0u){ //give the string time to go above threshold before trying to turn it off
                    if (curr_stringmax(y-1) < off_thres){ //If the vibration is less than 1/2 of the maximum vibration (or 1/4 for fretted notes), the note is considered done. Send MIDI off.
                        if (y < 3) {
                            sendMIDI(40u + 5*(6-y)+fret_active[y-1] - 1, OFF); //Map Note/Fret information into MIDI notes
                        }
                        else{
                            sendMIDI(40u + 5*(6-y)+fret_active[y-1], OFF);
                        }
                        string_active[y-1] = 0x0; //String is no longer active
                    }
                }
            }
         }
//////////////////////MIDI Note on Events
        //checked_string = Check_String(); //To improve performance for multi note situations
        if ((checked_string != 0u) & (prevbuffer2 == 0x0)) { //Only turn on once (equivalent to edge detection)
            thismidinote = 40u + 5*(6-checked_string)+Check_Fret(); //Map Fret/String information onto MIDI notes
            if (checked_string < 3) thismidinote = thismidinote-1;
            if ((string_active[checked_string-1]) ){ //First turn off any other notes on this string
                midi_off_note = 40u + 5*(6-checked_string)+fret_active[checked_string-1];
                if (checked_string < 3) midi_off_note = midi_off_note-1;
                sendMIDI(midi_off_note, OFF);
                }
            sendMIDI(thismidinote, ON);
            delay[checked_string-1] = 48u; //number of analog samples to delay by before comparing to thresholds. 48/6 = 8 samples per string.
            string_active[checked_string-1] = 0x1; //Keep track of what strings and frets are playing
            fret_active[checked_string-1] = Check_Fret();
         }

       
       prevbuffer2 = buffer[2];
       
///////////Display information for debugging     
       display_note = 6;
       LCD_Position(0,0);
       LCD_PrintString("Max  S");
       LCD_PrintInt8(display_note);
       LCD_PrintString(": ");
       LCD_PrintInt16(string_absmax[display_note-1]);
       LCD_Position(1,0);
       LCD_PrintString("Curr S");
       LCD_PrintInt8(display_note);
       LCD_PrintString(": ");
       LCD_PrintInt16(curr_stringmax(display_note-1));

       if(USB_IsConfigurationChanged() != 0u) /* Host could send double SET_INTERFACE request */
        {
            if(USB_GetConfiguration() != 0u)   /* Init IN endpoints when device configured */
            {
            	/* Enable the output endpoint */
                USB_MIDI_EP_Init();
            }  
        }        
 
}
}
/* [] END OF FILE */